import {
  Accordion,
  AccordionDetails,
  AccordionSummary,
} from '@material-ui/core'
import Button from '@material-ui/core/Button'
import chevronDown from 'assets/icons/ico/chevron-down.svg'
import CogIcon from 'assets/icons/ico/cog.svg'
import StyledIcon from 'components/CommonKit/Icon/StyledIcon'
import Loader from 'components/Loader/Loader'
import { useClient } from 'cozy-client'
import { useI18n } from 'cozy-ui/transpiled/react/providers/I18n'
import { FluidType } from 'enums'
import { MonthlyForecast, PerformanceIndicator } from 'models'
import React, { useCallback, useEffect, useState } from 'react'
import { useNavigate } from 'react-router-dom'
import ProfileTypeService from 'services/profileType.service'
import ProfileTypeEntityService from 'services/profileTypeEntity.service'
import { useAppSelector } from 'store/hooks'
import ProfileComparatorRow from './ProfileComparatorRow'
import './profileComparator.scss'

interface ProfileComparatorProps {
  aggregatedPerformanceIndicator: PerformanceIndicator
  performanceIndicators: PerformanceIndicator[]
}

const ProfileComparatorLayout = ({
  children,
}: {
  children: React.ReactNode
}) => {
  const { t } = useI18n()
  return (
    <div className="profileComparator">
      <div className="text-20-normal title">
        {t('analysis.comparison_title')}
      </div>
      {children}
    </div>
  )
}

const ProfileComparator = ({
  aggregatedPerformanceIndicator,
  performanceIndicators,
}: ProfileComparatorProps) => {
  const { t } = useI18n()
  const client = useClient()
  const navigate = useNavigate()
  const {
    analysis: { analysisMonth },
    profile,
  } = useAppSelector(state => state.ecolyo)
  const [homePriceConsumption, setHomePriceConsumption] = useState<number>(0)
  const [forecast, setForecast] = useState<MonthlyForecast | null>(null)
  const [isLoading, setIsLoading] = useState<boolean>(true)
  const [isExpanded, setIsExpanded] = useState<boolean>(false)

  const userPriceConsumption = aggregatedPerformanceIndicator.value || 0

  const emptyFluidTypes: FluidType[] = []
  if (performanceIndicators.length === 0) {
    // If no indicators add all fluids for component placeholder
    emptyFluidTypes.push(FluidType.ELECTRICITY, FluidType.WATER, FluidType.GAS)
  } else {
    for (let i = 0; i < performanceIndicators.length; i++) {
      if (!performanceIndicators[i]?.value) {
        emptyFluidTypes.push(i)
      }
    }
  }

  /** Toggle accordion and scroll to the first element */
  const toggleAccordion = () => {
    setIsExpanded(prev => !prev)
    if (!isExpanded) {
      setTimeout(() => {
        const content = document.querySelector('.consumption-electricity')
        if (content) {
          content.scrollIntoView({
            behavior: 'smooth',
            block: 'start',
          })
        }
      }, 300)
    }
  }

  const getTotalValueWithConnectedFluids = useCallback(
    (monthlyForecast: MonthlyForecast) => {
      if (performanceIndicators.length === 3) {
        setHomePriceConsumption(monthlyForecast.totalValue)
      } else {
        let totalPrice = 0
        performanceIndicators.forEach((fluid, index) => {
          if (monthlyForecast.fluidForecast[index].value)
            totalPrice += monthlyForecast.fluidForecast[index].value
        })
        setHomePriceConsumption(totalPrice)
      }
    },
    [performanceIndicators]
  )

  useEffect(() => {
    let subscribed = true
    async function loadAverageConsumption() {
      const analysisDate = analysisMonth.minus({ month: 1 }).startOf('month')
      const profileTypeEntityService = new ProfileTypeEntityService(client)
      const profileType =
        await profileTypeEntityService.getProfileType(analysisDate)
      if (profileType !== null) {
        const profileTypeService = new ProfileTypeService(
          profileType,
          client,
          analysisMonth.minus({ month: 1 }).year
        )
        const monthlyForecast = await profileTypeService.getMonthlyForecast(
          analysisDate.year,
          analysisDate.month
        )

        setForecast(monthlyForecast)
        if (monthlyForecast) {
          getTotalValueWithConnectedFluids(monthlyForecast)
        }
        setIsLoading(false)
      }
    }
    subscribed && loadAverageConsumption()
    return () => {
      subscribed = false
    }
  }, [
    profile.monthlyAnalysisDate.month,
    getTotalValueWithConnectedFluids,
    client,
    analysisMonth.month,
    analysisMonth,
  ])

  if (!profile.isProfileTypeCompleted) {
    return (
      <ProfileComparatorLayout>
        <div className="no-profile">
          <div className="text-16-normal">
            {t('analysis.approximative_description')}
          </div>
          <Button
            onClick={() => navigate('/profileType')}
            className="btnSecondary profileBtn"
            data-testid="goToProfile"
          >
            <StyledIcon icon={CogIcon} size={24} />
            {t('analysis.accessibility.button_go_to_profil')}
          </Button>
        </div>
      </ProfileComparatorLayout>
    )
  }

  if (isLoading) {
    return (
      <ProfileComparatorLayout>
        <div className="analysis-graph">
          <div className="loader-container">
            <Loader />
          </div>
        </div>
      </ProfileComparatorLayout>
    )
  }

  return (
    <ProfileComparatorLayout>
      <div className="analysis-graph">
        <div className="consumption-title text-14">
          <div className="user-title">{t('analysis.user_consumption')}</div>
          <div className="average-title">
            {t(`analysis.comparison_consumption`)}
          </div>
        </div>
        <ProfileComparatorRow
          fluidType={FluidType.MULTIFLUID}
          userPriceConsumption={userPriceConsumption}
          homePriceConsumption={homePriceConsumption}
          performanceValue={null}
          forecast={forecast}
          connected={performanceIndicators.length > 0}
          noData={false}
        />
        {performanceIndicators.map(
          (indicator, index) =>
            Boolean(indicator.value) && (
              <ProfileComparatorRow
                key={FluidType[index]}
                fluidType={index}
                userPriceConsumption={userPriceConsumption}
                homePriceConsumption={homePriceConsumption}
                performanceValue={indicator.value}
                forecast={forecast}
                connected={true}
                noData={false}
              />
            )
        )}
        {emptyFluidTypes.length > 0 && <hr className="consumption-sep" />}
        {emptyFluidTypes.map(fluid => (
          <ProfileComparatorRow
            key={fluid}
            fluidType={fluid}
            userPriceConsumption={userPriceConsumption}
            homePriceConsumption={homePriceConsumption}
            performanceValue={null}
            forecast={forecast}
            connected={false}
            noData={true}
          />
        ))}
        <Accordion
          expanded={isExpanded}
          onChange={toggleAccordion}
          classes={{
            root: 'expansion-panel-root',
          }}
        >
          <AccordionSummary
            aria-label={t(
              'profile_type.accessibility.button_toggle_average_home'
            )}
            expandIcon={
              <StyledIcon icon={chevronDown} className="accordion-icon" />
            }
            classes={{
              root: 'expansion-panel-summary',
              content: 'expansion-panel-content',
            }}
          >
            <div className="accordion-title accordion-title">
              {t('analysis.average_home')}
            </div>
          </AccordionSummary>
          <AccordionDetails
            classes={{
              root: 'expansion-panel-details',
            }}
          >
            <span className="accordion-desc text-16-normal">
              {t('analysis.average_home_description')}
            </span>
          </AccordionDetails>
        </Accordion>
        <Button
          onClick={() => navigate('/profileType')}
          className="btnSecondary profileBtn"
          data-testid="iconGoToProfile"
        >
          <StyledIcon icon={CogIcon} size={24} />
          {t('analysis.average_edit_profile')}
        </Button>
      </div>
    </ProfileComparatorLayout>
  )
}

export default ProfileComparator
