import * as Sentry from '@sentry/react'
import { Client } from 'cozy-client'
import logger from 'cozy-logger'
import {
  FluidState,
  FluidType,
  HotWaterFluid,
  TimeStep,
  UsageEventType,
  WarmingType,
} from 'enums'
import { uniq } from 'lodash'
import { DateTime } from 'luxon'
import { PerformanceIndicator, UsageEvent } from 'models'
import ConsumptionService from 'services/consumption.service'
import EnvironmentService from 'services/environment.service'
import FluidService from 'services/fluid.service'
import ProfileService from 'services/profile.service'
import ProfileTypeEntityService from 'services/profileTypeEntity.service'
import TermsService from 'services/terms.service'
import UsageEventService from 'services/usageEvent.service'
import { getFluidName } from 'utils/utils'
import { runService } from './service'

const logStack = logger.namespace('aggregatorUsageEvents')

enum DaccIndicator {
  CONSUMPTION_VARIATION_MONTHLY = 'consumption-variation-monthly',
  FLUID_DATA_GRANULARITY = 'fluid-data-granularity-monthly',
  KONNECTOR_CONNECTED_PER_DAY = 'konnector-connected-daily',
  PROFILE_COUNT_MONTHLY = 'profile-count',
  SUMMARY_SUBSCRIPTION_MONTHLY = 'summary-subscription-monthly',
}

interface AggregatorUsageEventsProps {
  client: Client
}

interface Indicator {
  createdBy: string
  measureName: string
  startDate: string
  value: number | null
  group1?: object
  group2?: object
  group3?: object
}

// Store all id of events read during the process
const readUsageEvents: UsageEvent[] = []

// Store error events
const errorEvent: { error: unknown; doctype: string }[] = []

const sendIndicator = async (
  indicator: Indicator,
  client: Client
): Promise<boolean> => {
  try {
    const environmentService = new EnvironmentService()
    logStack(
      'info',
      environmentService.isProduction()
        ? 'Sending data to dacc'
        : 'Sending data to dacc-dev'
    )

    // /!\ In order to test locally, please replace /remote/cc.cozycloud.dacc with http://localhost:8081
    await client
      .getStackClient()
      .fetchJSON(
        'POST',
        environmentService.isProduction()
          ? '/remote/cc.cozycloud.dacc_v2'
          : '/remote/cc.cozycloud.dacc.dev_v2',
        {
          data: JSON.stringify(indicator),
          path: 'measure',
        }
      )
    return true
  } catch (error) {
    let errorMessage = 'Failed to do something exceptional'
    if (error instanceof Error) {
      errorMessage = error.message
    }
    logStack(
      'error',
      `Error while sending indicator ${indicator.measureName} to remote doctype: ${errorMessage}`
    )
    Sentry.captureException(error)
    throw error
  }
}

/**
 * Format a string with all konnectors in success state
 * @param konnectorFluidTypes FluidType[]
 * @returns string
 */
const getConnectedKonnectorSlug = (
  konnectorFluidTypes: FluidType[]
): string => {
  let slug = ''
  if (konnectorFluidTypes.includes(FluidType.ELECTRICITY)) {
    slug += 'electricity'
  }
  if (konnectorFluidTypes.includes(FluidType.GAS)) {
    if (slug.length > 0) {
      slug += ':'
    }
    slug += 'gas'
  }
  if (konnectorFluidTypes.includes(FluidType.WATER)) {
    if (slug.length > 0) {
      slug += ':'
    }
    slug += 'water'
  }
  return slug
}

const calculateConnectedKonnectorPerDay = async (client: Client) => {
  const fluidService = new FluidService(client)
  const fluidStatus = await fluidService.getFluidStatus()
  const connectedKonnectors = fluidStatus.filter(
    fluid => fluid.status === FluidState.DONE
  )
  logStack('info', 'calculateConnectedKonnectorPerDay')
  if (connectedKonnectors.length > 0) {
    const konnectorFluidTypes: FluidType[] = []
    for (const konnector of connectedKonnectors) {
      konnectorFluidTypes.push(konnector.fluidType)
    }
    const KonnectorConnectedPerDayIndicator: Indicator = {
      createdBy: 'ecolyo',
      measureName: DaccIndicator.KONNECTOR_CONNECTED_PER_DAY,
      startDate: DateTime.local()
        .setZone('utc', {
          keepLocalTime: true,
        })
        .startOf('day')
        .toISODate(),
      value: connectedKonnectors.length,
      group1: { categories: getConnectedKonnectorSlug(konnectorFluidTypes) },
    }
    await sendIndicator(KonnectorConnectedPerDayIndicator, client)
  }
}

/**
 * Build indicator group string with handling of : ECS, heating for gas and electricity.
 * @param client Client
 * @param fluidType FluidType
 * @returns Promise<string>
 */
const buildProfileWithFluidType = async (
  client: Client,
  fluidType: FluidType
): Promise<string> => {
  let formattedProfile = ''
  const profile = await new ProfileService(client).getProfile()
  // If profile is not filled, return empty string

  if (!profile?.isProfileTypeCompleted) return formattedProfile
  const date = DateTime.local().setZone('utc', {
    keepLocalTime: true,
  })
  const profileType = await new ProfileTypeEntityService(client).getProfileType(
    date
  )
  if (fluidType === FluidType.ELECTRICITY) {
    if (profile && profileType?.warmingFluid === WarmingType.ELECTRICITY) {
      formattedProfile = 'chauffage élec'
    }
    if (profile && profileType?.warmingFluid === WarmingType.WOOD) {
      formattedProfile = 'chauffage bois'
    }
    if (profile && profileType?.warmingFluid === WarmingType.FUEL) {
      formattedProfile = 'chauffage fuel'
    }
    if (profile && profileType?.hotWaterFluid === HotWaterFluid.ELECTRICITY) {
      if (formattedProfile.length === 0) {
        formattedProfile = 'ECS'
      } else {
        formattedProfile += ':ECS'
      }
    }
    if (profile && profileType?.cookingFluid === FluidType.ELECTRICITY) {
      if (formattedProfile.length === 0) {
        formattedProfile = 'cuisine élec'
      } else {
        formattedProfile += ':cuisine élec'
      }
    }
    if (formattedProfile.length === 0) {
      return 'électricité spécifique'
    } else {
      return formattedProfile
    }
  } else if (fluidType === FluidType.GAS) {
    if (profile && profileType?.warmingFluid === WarmingType.GAS) {
      formattedProfile = 'chauffage gaz'
    }
    if (
      profile &&
      profileType &&
      profileType.hotWaterFluid === HotWaterFluid.GAS
    ) {
      if (formattedProfile.length === 0) {
        formattedProfile = 'ECS'
      } else {
        formattedProfile += ':ECS'
      }
    }
    if (profile && profileType && profileType.cookingFluid === FluidType.GAS) {
      if (formattedProfile.length === 0) {
        formattedProfile = 'cuisine gaz'
      } else {
        formattedProfile += ':cuisine gaz'
      }
    }
    if (formattedProfile.length === 0) {
      return 'autre'
    } else {
      return formattedProfile
    }
  }
  return formattedProfile
}

const getConsumptionValue = async (
  client: Client,
  fluidTypes: FluidType[]
): Promise<PerformanceIndicator[]> => {
  const consumptionService = new ConsumptionService(client)
  const analysisDate = DateTime.local().setZone('utc', { keepLocalTime: true })
  const periods = {
    timePeriod: {
      startDate: analysisDate.minus({ month: 1 }).startOf('month'),
      endDate: analysisDate.minus({ month: 1 }).endOf('month'),
    },
    comparisonTimePeriod: {
      startDate: analysisDate.minus({ month: 2 }).startOf('month'),
      endDate: analysisDate.minus({ month: 2 }).endOf('month'),
    },
  }
  const fetchedPerformanceIndicators =
    await consumptionService.getPerformanceIndicators(
      periods.timePeriod,
      TimeStep.MONTH,
      fluidTypes,
      periods.comparisonTimePeriod
    )
  return fetchedPerformanceIndicators
}

/**
 * Send an indicator on the consumption variation in % for each fluid type.
 * @description [{ slug }, { seniority (in month) }, { profile (ECS, chauffage, etc...) }],
 */
const calculateConsumptionVariation = async (client: Client) => {
  logStack('info', `calculateConsumptionVariation`)
  const consumptionData = await getConsumptionValue(client, [
    FluidType.ELECTRICITY,
    FluidType.GAS,
    FluidType.WATER,
  ])
  for (const fluidType of [
    FluidType.ELECTRICITY,
    FluidType.GAS,
    FluidType.WATER,
  ]) {
    // Seniority is set to 0, otherwise the indicator is too specific
    const seniority = 0

    const consumptionVariationIndicator: Indicator = {
      createdBy: 'ecolyo',
      measureName: DaccIndicator.CONSUMPTION_VARIATION_MONTHLY,
      startDate: DateTime.local()
        .setZone('utc', {
          keepLocalTime: true,
        })
        .startOf('day')
        .toISODate(),
      value: consumptionData[fluidType]?.percentageVariation
        ? consumptionData[fluidType].percentageVariation
        : 0, // in percent
      // eslint-disable-next-line camelcase
      group1: { fluid_type: getFluidName(fluidType) },
      group2: { seniority: Math.round(seniority).toString() },
      group3: {
        // eslint-disable-next-line camelcase
        fluid_usage: await buildProfileWithFluidType(client, fluidType),
      },
    }

    // if user wasn't connected during current month, dont send indicator
    const events = await UsageEventService.getEvents(client, {
      eventDate: {
        $lt: DateTime.local()
          .setZone('utc', {
            keepLocalTime: true,
          })
          .endOf('month')
          .minus({ month: 1 })
          .toString(),
        $gt: DateTime.local()
          .setZone('utc', {
            keepLocalTime: true,
          })
          .startOf('month')
          .minus({ month: 1 })
          .toString(),
      },
    })

    if (events.length > 0 && consumptionVariationIndicator.value !== 0) {
      logStack(
        'info',
        `Send variation indicator : ${JSON.stringify(
          consumptionVariationIndicator
        )}`
      )
      await sendIndicator(consumptionVariationIndicator, client)
    }
  }
}

const sendProfileCount = async (client: Client) => {
  logStack('info', `sendProfileCount`)
  // Get profile setEvents
  const events = await UsageEventService.getEvents(client, {
    type: UsageEventType.PROFILE_SET_EVENT,
  })

  // If there is at least one connection, filter each unique connection in order to send it
  if (events.length > 0) {
    const profileSet: Indicator = {
      createdBy: 'ecolyo',
      measureName: DaccIndicator.PROFILE_COUNT_MONTHLY,
      startDate: DateTime.local()
        .setZone('utc', {
          keepLocalTime: true,
        })
        .startOf('day')
        .toISODate(),
      value: events.length,
    }
    await sendIndicator(profileSet, client)
  }
}

/** Send if user has came back to Ecolyo using the button in the newsletter */
const sendNewsletterComeBack = async (client: Client) => {
  logStack('info', `sendNewsletterComeBack`)
  const events = await UsageEventService.getEvents(client, {
    type: UsageEventType.REPORT_FROM_EMAIL,
    eventDate: {
      $lt: DateTime.local()
        .setZone('utc', {
          keepLocalTime: true,
        })
        .minus({ days: 1 })
        .endOf('day')
        .toString(),
      $gt: DateTime.local()
        .setZone('utc', {
          keepLocalTime: true,
        })
        .minus({ days: 1 })
        .startOf('day')
        .toString(),
    },
  })
  if (events.length) {
    const cameBackFromNewsletter: Indicator = {
      createdBy: 'ecolyo',
      measureName: DaccIndicator.SUMMARY_SUBSCRIPTION_MONTHLY,
      startDate: DateTime.local()
        .setZone('utc', {
          keepLocalTime: true,
        })
        .minus({ days: 1 })
        .startOf('day')
        .toISODate(),
      value: 1,
      // eslint-disable-next-line camelcase
      group1: { event_type: 'email' },
    }
    await sendIndicator(cameBackFromNewsletter, client)
  }
}

const sendNewsletterSubscription = async (client: Client) => {
  logStack('info', `sendEmailSubscription`)
  const profile = await new ProfileService(client).getProfile()
  if (!profile) return
  const isSubscribed: Indicator = {
    createdBy: 'ecolyo',
    measureName: DaccIndicator.SUMMARY_SUBSCRIPTION_MONTHLY,
    startDate: DateTime.local()
      .setZone('utc', {
        keepLocalTime: true,
      })
      .startOf('day')
      .toISODate(),
    value: profile.sendAnalysisNotification ? 1 : 0,
    // eslint-disable-next-line camelcase
    group1: { event_type: 'subscribed' },
  }
  await sendIndicator(isSubscribed, client)
}

/**
 * Send indicator to dacc, if user has half-hour data.
 * @param client CozyClient
 */
const sendHalfHourConsumption = async (client: Client) => {
  logStack('info', `sendHalfHourConsumption`)
  const consumptionService = new ConsumptionService(client)

  const data = await consumptionService.getLastHourData(
    client,
    DateTime.local().minus({ month: 1 }).startOf('month').month
  )

  const halfHourConsumption: Indicator = {
    createdBy: 'ecolyo',
    measureName: DaccIndicator.FLUID_DATA_GRANULARITY,
    // eslint-disable-next-line camelcase
    group1: { fluid_type: 'electricity' },
    group2: { granularity: 'half_hour' },
    startDate: DateTime.local()
      .setZone('utc', {
        keepLocalTime: true,
      })
      .startOf('day')
      .toISODate(),
    value: data?.length > 0 ? 1 : 0,
  }
  await sendIndicator(halfHourConsumption, client)
}

const AggregatorUsageEvents = async ({
  client,
}: AggregatorUsageEventsProps) => {
  logStack('info', 'Launch service')
  logStack('info', 'Dacc consent validation...')
  const termService = new TermsService(client)
  if (!(await termService.getLastTerm())) {
    logStack('info', 'Exit: no consent available')
    return
  }

  // Daily events
  await calculateConnectedKonnectorPerDay(client)
  await sendNewsletterComeBack(client)
  // Monthly events
  const profile = await new ProfileService(client).getProfile()
  if (
    profile &&
    DateTime.local()
      .setZone('utc', {
        keepLocalTime: true,
      })
      .startOf('day').day === profile.monthlyAnalysisDate.day
  ) {
    await calculateConsumptionVariation(client)
    await sendNewsletterSubscription(client)
    await sendHalfHourConsumption(client)
    await sendProfileCount(client)
  }

  const uniqueReadUsageEvents: UsageEvent[] = uniq(readUsageEvents)
  logStack(
    'info',
    `Tag aggregated usage: total of ${uniqueReadUsageEvents.length} events`
  )
  await UsageEventService.updateUsageEventsAggregated(
    client,
    uniqueReadUsageEvents
  )

  if (errorEvent.length > 0) {
    // Build error
    let error = `Process end with ${errorEvent.length} errors : `
    errorEvent.forEach(el => {
      error += `${el.doctype}, `
    })
    logStack('error', error)
    Sentry.captureException(error)
    throw error
  }

  logStack('info', 'End of service')
}

runService(AggregatorUsageEvents)
